/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Copyright (C) 2004 
 *
 * Authors: Erik Eliasson <eliasson@it.kth.se>
 *          Johan Bilien <jobi@via.ecp.fr>
*/

#include<libmikey/MikeyMessageDH.h>
#include<libmikey/MikeyException.h>

MikeyMessageDH::MikeyMessageDH( KeyAgreementDH * ka, unsigned int ssrc, unsigned int roc ):MikeyMessage(){

	certificate_chain * certChain;
	certificate * cert;
	/* generate random a CryptoSessionBundle ID */
	unsigned int csbId = rand();
	ka->setCsbId( csbId );
	MikeyCsIdMapSrtp * csIdMap = new MikeyCsIdMapSrtp();
	csIdMap->addCs( 0, ssrc, roc );

	addPayload( 
		new MikeyPayloadHDR( HDR_DATA_TYPE_DH_INIT, 1, 
			HDR_PRF_MIKEY_1, csbId,
			1, HDR_CS_ID_MAP_TYPE_SRTP_ID, 
			csIdMap ) );

	addPayload( new MikeyPayloadT() );

	MikeyPayloadRAND * payload;
	addPayload( payload = new MikeyPayloadRAND() );
	
	//keep a copy of the random value!
	ka->setRand( payload->randData(), 
		     payload->randLength() );

	/* Include the list of certificates if available */
	if( ( certChain = ka->certificateChain() ) != NULL ){
		ka->certificateChain()->lock();
		while( ( cert = certChain->get_next() ) != NULL ){
			addPayload( new MikeyPayloadCERT(
				MIKEYPAYLOAD_CERT_TYPE_X509V3SIGN,
				cert) );
		}
		ka->certificateChain()->unlock();
	}

	addPayload( new MikeyPayloadDH( ka->group(),
					ka->publicKey(),
					ka->keyValidity() ) );

	addSignaturePayload( ka->certificateChain()->get_first() );

}

MikeyMessage * MikeyMessageDH::buildResponse( KeyAgreementDH * ka ){

	MikeyPayload * i = extractPayload( MIKEYPAYLOAD_HDR_PAYLOAD_TYPE );
	bool error = false;
	MikeyCsIdMap * csIdMap;
	certificate * peerCert = NULL;
	MikeyMessage * errorMessage = new MikeyMessage();
	certificate * cert;
	certificate_chain * certChain;

	if( i == NULL ){
		throw new MikeyExceptionMessageContent(
				"DH init message had no HDR payload" );
	}

#define hdr ((MikeyPayloadHDR *)i)
	if( hdr->dataType() != HDR_DATA_TYPE_DH_INIT )
		throw new MikeyExceptionMessageContent( 
				"Expected DH init message" );

	unsigned int csbId = hdr->csbId();
	ka->setCsbId( csbId );

	if( hdr->csIdMapType() == HDR_CS_ID_MAP_TYPE_SRTP_ID ){
		csIdMap = hdr->csIdMap();
	}
	else{
		throw new MikeyExceptionMessageContent( 
				"Unknown type of CS ID map" );
	}
	payloads.remove( i );
#undef hdr

	//FIXME look at the other fields!

	errorMessage->addPayload(
			new MikeyPayloadHDR( HDR_DATA_TYPE_ERROR, 0,
			HDR_PRF_MIKEY_1, csbId,
			1, HDR_CS_ID_MAP_TYPE_SRTP_ID, 
			csIdMap ) );
	
	i = extractPayload( MIKEYPAYLOAD_T_PAYLOAD_TYPE );

	if( i == NULL ){
		error = true;
		errorMessage->addPayload( 
			new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
	}

	if( ((MikeyPayloadT*)i)->checkOffset( MAX_TIME_OFFSET ) ){
		error = true;
		errorMessage->addPayload( 
			new MikeyPayloadERR( MIKEY_ERR_TYPE_INVALID_TS ) );
	}
	
	payloads.remove( i );
	i = extractPayload( MIKEYPAYLOAD_RAND_PAYLOAD_TYPE );

	if( i == NULL ){
		error = true;
		errorMessage->addPayload( 
			new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
	}

	ka->setRand( ((MikeyPayloadRAND *)i)->randData(),
		     ((MikeyPayloadRAND *)i)->randLength() );

	payloads.remove( i );

	/* If we haven't gotten the peer's certificate chain
	 * (for instance during authentication of the message),
	 * try to get it now */

	if( ka->peerCertificateChain()->get_first() == NULL ){
		i = extractPayload( MIKEYPAYLOAD_CERT_PAYLOAD_TYPE );

		while( i != NULL )
		{
			peerCert = new certificate( 
				((MikeyPayloadCERT *)i)->certData(),
				((MikeyPayloadCERT *)i)->certLength()
				);

			ka->addPeerCertificate( peerCert );
			payloads.remove( i );

			i = extractPayload( MIKEYPAYLOAD_CERT_PAYLOAD_TYPE );
		}
	}

	//FIXME treat the case of an ID payload
	/*
	{
		i = extractPayload( MIKEYPAYLOAD_ID_PAYLOAD_TYPE );
		if( i != NULL ){
			payloads.remove( i );
		}
	}
	*/


	i = extractPayload( MIKEYPAYLOAD_DH_PAYLOAD_TYPE );

	if( i == NULL ){
		error = true;
		errorMessage->addPayload( 
			new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
	}


	if( ka->group() != ((MikeyPayloadDH *)i)->group() ){
		ka->setGroup( ((MikeyPayloadDH *)i)->group() );
	}

	ka->setPeerKey( ((MikeyPayloadDH *)i)->dhKey(),
		        ((MikeyPayloadDH *)i)->dhKeyLength() );
	
	ka->setKeyValidity( ((MikeyPayloadDH *)i)->kv() );
	
	payloads.remove( i );

	// Build the response message
	MikeyMessage * result = new MikeyMessage();
	result->addPayload( 
			new MikeyPayloadHDR( HDR_DATA_TYPE_DH_RESP, 0, 
			HDR_PRF_MIKEY_1, csbId,
			1, HDR_CS_ID_MAP_TYPE_SRTP_ID, 
			csIdMap ) );

	result->addPayload( new MikeyPayloadT() );

	/* Include the list of certificates if available */
	if( ( certChain = ka->certificateChain() ) != NULL ){
		ka->certificateChain()->lock();
		while( ( cert = certChain->get_next() ) != NULL ){
			result->addPayload( new MikeyPayloadCERT(
				MIKEYPAYLOAD_CERT_TYPE_X509V3SIGN,
				cert) );
		}
		ka->certificateChain()->unlock();
	}

	result->addPayload( new MikeyPayloadDH( 
				    ka->group(),
                                    ka->publicKey(),
				    ka->keyValidity() ) );

	result->addPayload( new MikeyPayloadDH( 
				    ka->group(),
                                    ka->peerKey(),
				    ka->keyValidity() ) );

	result->addSignaturePayload( ka->certificateChain()->get_first() );

	return result;
}

MikeyMessage * MikeyMessageDH::parseResponse( KeyAgreementDH * ka ){
	MikeyPayload * i = extractPayload( MIKEYPAYLOAD_HDR_PAYLOAD_TYPE );
	bool error = false;
	bool gotDhi = false;
	certificate * peerCert;
	MikeyMessage * errorMessage = new MikeyMessage();
	MikeyCsIdMap * csIdMap;
	
	if( i == NULL ){
		throw new MikeyExceptionMessageContent(
				"DH resp message had no HDR payload" );
	}

#define hdr ((MikeyPayloadHDR *)(i))
	if( hdr->dataType() != HDR_DATA_TYPE_DH_RESP ){
		throw new MikeyExceptionMessageContent( 
				"Expected DH resp message" );
	}

	if( hdr->csIdMapType() == HDR_CS_ID_MAP_TYPE_SRTP_ID ){
		csIdMap = hdr->csIdMap();
	}
	else{
		throw new MikeyExceptionMessageContent( 
				"Unknown type of CS ID map" );
	}
#undef hdr
	//FIXME look at the other fields!
	
	errorMessage->addPayload(
			new MikeyPayloadHDR( HDR_DATA_TYPE_ERROR, 0,
			HDR_PRF_MIKEY_1, ka->csbId(),
			1, HDR_CS_ID_MAP_TYPE_SRTP_ID,
			csIdMap ) );
	
	payloads.remove( i );
	i = extractPayload( MIKEYPAYLOAD_T_PAYLOAD_TYPE );

	if( i == NULL ){
		error = true;
		errorMessage->addPayload( 
			new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
	}

	if( ((MikeyPayloadT*)i)->checkOffset( MAX_TIME_OFFSET ) ){
		error = true;
		errorMessage->addPayload( 
			new MikeyPayloadERR( MIKEY_ERR_TYPE_INVALID_TS ) );
	}

	payloads.remove( i );
	
	if( ka->peerCertificateChain()->get_first() == NULL ){
		i = extractPayload( MIKEYPAYLOAD_CERT_PAYLOAD_TYPE );

		while( i != NULL )
		{
			peerCert = new certificate( 
				((MikeyPayloadCERT *)i)->certData(),
				((MikeyPayloadCERT *)i)->certLength()
				);

			ka->addPeerCertificate( peerCert );
			payloads.remove( i );

			i = extractPayload( MIKEYPAYLOAD_CERT_PAYLOAD_TYPE );
		}
	}

	//FIXME treat the case of an ID payload
	/*{
		i = extractPayload( MIKEYPAYLOAD_ID_PAYLOAD_TYPE );
		if( i != NULL ){
			payloads.remove( i );
		}
	}*/

	i = extractPayload( MIKEYPAYLOAD_DH_PAYLOAD_TYPE );
	
	if( i == NULL ){
		error = true;
		errorMessage->addPayload(
			new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
	}

#define dh ((MikeyPayloadDH *)i)
	if( string( (const char *)dh->dhKey(), 
				  dh->dhKeyLength() ) ==
	    string( (const char *)ka->publicKey(), 
		    		  ka->publicKeyLength() ) ){
		// This is the DHi payload
		gotDhi = true;
	}
	else{
		// This is the DHr payload
		ka->setPeerKey( dh->dhKey(),
				dh->dhKeyLength() );
	}

	payloads.remove( i );
	i = extractPayload( MIKEYPAYLOAD_DH_PAYLOAD_TYPE );

	if( i == NULL ){
		error = true;
		errorMessage->addPayload(
			new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
	}

	if( gotDhi ){
		// this one should then be DHr
		ka->setPeerKey( dh->dhKey(), dh->dhKeyLength() );
	}
	else{
		if( string( (const char *)dh->dhKey(), 
					  dh->dhKeyLength() ) !=
	    	    string( (const char *)ka->publicKey(), 
		    	    ka->publicKeyLength() ) ){
			error = true;
			errorMessage->addPayload(
				new MikeyPayloadERR( MIKEY_ERR_TYPE_UNSPEC ) );
		}
	}
#undef dh

        //FIXME handle key validity information
	
	if( error ){
		cerr << "An Error was found in the MIKEY Response message" << endl; 
		errorMessage->addSignaturePayload( 
				ka->certificateChain()->get_first() );
		throw new MikeyExceptionMessageContent( errorMessage );
	}

	delete errorMessage;

	return NULL;
}

bool MikeyMessageDH::authenticate( KeyAgreement * _ka ){
	KeyAgreementDH * ka = (KeyAgreementDH *)ka;

	MikeyPayload * sign = (*lastPayload());
	list<MikeyPayload *>::iterator i;
	certificate_chain * peerCert = ka->peerCertificateChain();

	if( peerCert == NULL || peerCert->get_first() == NULL )
	{
		/* Try to find the certificate chain in the message */
		MikeyPayloadCERT * certPayload;
		while( ( ((MikeyPayload *)certPayload) = 
				extractPayload( MIKEYPAYLOAD_CERT_PAYLOAD_TYPE)
				) != NULL ){
			ka->addPeerCertificate(
				new certificate( 
					certPayload->certData(),
					certPayload->certLength() ));
			payloads.remove( certPayload );
		}
	}

	if( peerCert->get_first() == NULL ){
		ka->setAuthError( "No certificate was found" );
		return true;
	}


	if( sign->payloadType() != MIKEYPAYLOAD_SIGN_PAYLOAD_TYPE ){
		ka->setAuthError( "No signature payload found" );
		return true;
	}

	return peerCert->get_first()->verif_sign( rawMessageData(),
			rawMessageLength() - 
			((MikeyPayloadSIGN *)sign)->sigLength(),
			((MikeyPayloadSIGN *)sign)->sigData(),
			((MikeyPayloadSIGN *)sign)->sigLength() );

}
