/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Copyright (C) 2004 
 *
 * Authors: Erik Eliasson <eliasson@it.kth.se>
 *          Johan Bilien <jobi@via.ecp.fr>
*/

#include<libmikey/MikeyPayloadHDR.h>
#include<libmikey/MikeyException.h>
#include<iostream>
#include<assert.h>


MikeyPayloadHDR::MikeyPayloadHDR( int dataType, int v, int prfFunc, int csbId, 
		int nCs, int mapType, MikeyCsIdMap * map ) {

	this->payloadTypeValue = MIKEYPAYLOAD_HDR_PAYLOAD_TYPE;
	this->version = 1;
	this->dataTypeValue = dataType;
	this->vValue = v;
	this->nCs = nCs;
	this->prfFunc = prfFunc;
	this->csbIdValue = csbId;
	if( mapType == HDR_CS_ID_MAP_TYPE_SRTP_ID ){
		this->csIdMapTypeValue = mapType;
		this->csIdMapPtr = map;
	}
	else{
		throw new MikeyExceptionMessageContent(
				"Unknown CS ID map type" );
	}
	
}

MikeyPayloadHDR::MikeyPayloadHDR( byte_t * start, int lengthLimit ):
	MikeyPayload( start ){

	this->payloadTypeValue = MIKEYPAYLOAD_HDR_PAYLOAD_TYPE;
	if( lengthLimit < 10 ){
		throw new MikeyExceptionMessageLengthException(
			"Given data is too short to form a HDR Payload" );
		return;
	}

	setNextPayloadType( start[2] );
	this->version = start[0];
	this->dataTypeValue = start[1];
	this->vValue = ( start[3] >> 7 ) & 0x1;
	this->prfFunc = start[3] & 0x7F;
	this->csbIdValue = (int)start[4] << 24 |
		           (int)start[5] << 16 |
		           (int)start[6] <<  8 |
		           (int)start[7];
	this->nCs = start[8];
	this->csIdMapTypeValue = start[9];
	if( csIdMapTypeValue == HDR_CS_ID_MAP_TYPE_SRTP_ID ){
		if( lengthLimit < 10 + nCs * 9 ){
			throw new MikeyExceptionMessageLengthException(
			"Given data is too short to form a HDR Payload" );
			return;
		}
		this->csIdMapPtr = 
		   new MikeyCsIdMapSrtp( &start[10], 9 * nCs );
		this->endPtr = startPtr + 10 + 9 * this->nCs;
	}
	else{
		throw new MikeyExceptionMessageContent( 
				"Unknown type of CS_ID_map" );
		this->csIdMapPtr = NULL;
	}
}

MikeyPayloadHDR::~MikeyPayloadHDR(){
	if( csIdMapPtr )
		delete csIdMapPtr;
}
	
int MikeyPayloadHDR::length(){
	return 10 + csIdMapPtr->length();
}

void MikeyPayloadHDR::writeData( byte_t * start, int expectedLength ){
	assert( expectedLength == length() );
	memset( start, expectedLength, 0 );
	start[0] = (byte_t) version;
	start[1] = (byte_t) dataTypeValue;
	start[2] = nextPayloadType();
	start[3] = ( (byte_t)vValue & 0x1) << 7 
		 | ( (byte_t)prfFunc & 0x7F );
	start[4] = (byte_t) ((csbIdValue & 0xFF000000) >> 24);
	start[5] = (byte_t) ((csbIdValue & 0xFF0000) >> 16);
	start[6] = (byte_t) ((csbIdValue & 0xFF00) >> 8);
	start[7] = (byte_t) (csbIdValue & 0xFF);
	start[8] = (byte_t) nCs;
	start[9] = (byte_t) csIdMapTypeValue;
	csIdMapPtr->writeData( &start[10], csIdMapPtr->length() );

}

string MikeyPayloadHDR::debugDump(){
	string ret= "MikeyPayloadHDR: version=<"+itoa(version)+"> datatype=";
	switch( dataTypeValue ){
		case  HDR_DATA_TYPE_PSK_INIT: 
			ret=ret+"<Pre-shared>";
			break;
		case HDR_DATA_TYPE_PSK_RESP:
			ret=ret+"<PS ver msg>";
			break;
		case HDR_DATA_TYPE_PK_INIT:
			ret=ret+"<Public key>";
			break;
		case HDR_DATA_TYPE_PK_RESP:
			ret=ret+"<PK ver msg>";
			break;
		case  HDR_DATA_TYPE_DH_INIT:
			ret=ret+"<D-H init>";
			break;
		case HDR_DATA_TYPE_DH_RESP:
			ret=ret+"<D-H resp>";
			break;
		case HDR_DATA_TYPE_ERROR:
			ret=ret+"<Error>";
			break;
	}

	ret += " next_payload=" + itoa( nextPayloadType() );
	ret += " V=" + itoa( vValue );
	ret += " PRF_func=";
	switch( prfFunc ){
		case HDR_PRF_MIKEY_1:
			ret += "<MIKEY-1>";
			break;
		case  HDR_PRF_MIKEY_256:
			ret += "<MIKEY-256>";
			break;
		case HDR_PRF_MIKEY_384:
			ret += "<MIKEY-384>";
			break;
		case  HDR_PRF_MIKEY_512:
			ret += "<MIKEY-512>";
			break;
	}
	
	ret += " CSB_id=<" + itoa( csbIdValue ) + ">";
	ret += " #CS=<" + itoa( nCs ) + ">";
	ret += " CS ID map type=";
	if ( csIdMapTypeValue == HDR_CS_ID_MAP_TYPE_SRTP_ID )
		ret += "<SRTP-ID>";
	else
		ret += "<unknown (" + itoa( csIdMapTypeValue ) + ")>";
	
	return ret;
}

int MikeyPayloadHDR::dataType(){
	return dataTypeValue;
}

int MikeyPayloadHDR::v(){
	return vValue;
}

unsigned int MikeyPayloadHDR::csbId(){
	return csbIdValue;
}

int MikeyPayloadHDR::csIdMapType(){
	return csIdMapTypeValue;
}

MikeyCsIdMap * MikeyPayloadHDR::csIdMap(){
	return csIdMapPtr;
}
MikeySrtpCs::MikeySrtpCs( int policyNo, int ssrc, int roc ):
	policyNo( policyNo ), ssrc( ssrc ), roc( roc ){};

MikeyCsIdMapSrtp::MikeyCsIdMapSrtp(){
	cs = list<MikeySrtpCs *>::list();
}

MikeyCsIdMapSrtp::MikeyCsIdMapSrtp( byte_t * data, int length ){
	if( length % 9 ){
		throw new MikeyException( 
				"Invalid length of SRTP_ID map info" );
	}

	int nCs = length / 9;
	int i, policyNo, ssrc, roc;

	for( i = 0; i < nCs; i++ ){
		policyNo = (int)data[ i*9 ];
		ssrc = (int)data[ i*9 + 1 ] << 24 |
		       (int)data[ i*9 + 2 ] << 16 |
		       (int)data[ i*9 + 3 ] <<  8 |
		       (int)data[ i*9 + 4 ];
		roc  = (int)data[ i*9 + 5 ] << 24 |
		       (int)data[ i*9 + 6 ] << 16 |
		       (int)data[ i*9 + 7 ] <<  8 |
		       (int)data[ i*9 + 8 ];
		addCs( policyNo, ssrc, roc );
	}
}



MikeyCsIdMapSrtp::~MikeyCsIdMapSrtp(){
	list<MikeySrtpCs *>::iterator i;

	for( i = cs.begin(); i!= cs.end() ; i++ )
		delete *i;
}

int MikeyCsIdMapSrtp::length(){
	return 9 * cs.size();
}

void MikeyCsIdMapSrtp::writeData( byte_t * start, int expectedLength ){
	if( expectedLength < length() ){
		throw new MikeyExceptionMessageLengthException(
				"CsSrtpId is too long" );
	}

	int j = 0,k;
	list<MikeySrtpCs *>::iterator i;

	for( i = cs.begin(); i != cs.end(); i++ ){
		start[ 9*j ] = (*i)->policyNo & 0xFF;
		for( k = 0; k < 4; k++ ){
			start[9*j+1+k] = ((*i)->ssrc >> (3-j)) & 0xFF;
		}
		for( k = 0; k < 4; k++ ){
			start[9*j+5+k] = ((*i)->roc >> (3-j)) & 0xFF;
		}
		j++;
	}
}

void MikeyCsIdMapSrtp::addCs( int policy, int ssrc, int roc ){
	cs.push_back( new MikeySrtpCs( policy, ssrc, roc ) );
}

