/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Copyright (C) 2004 
 *
 * Authors: Erik Eliasson <eliasson@it.kth.se>
 *          Johan Bilien <jobi@via.ecp.fr>
*/

#include<libmikey/MikeyPayloadKEMAC.h>
#include<libmikey/MikeyException.h>
#include<iostream>
#include<assert.h>

#include"../util/aes.h"

MikeyPayloadKEMAC::MikeyPayloadKEMAC( int encrAlgValue, 
		int encrDataLengthValue, byte_t * encrDataPtr, 
		int macAlgValue, byte_t * macDataPtr ){
	this->payloadTypeValue = MIKEYPAYLOAD_KEMAC_PAYLOAD_TYPE;
	this->encrAlgValue = encrAlgValue;
	this->encrDataLengthValue = encrDataLengthValue;
	this->encrDataPtr = new byte_t[ encrDataLengthValue ];
	memcpy( this->encrDataPtr, encrDataPtr, encrDataLengthValue );
	this->macAlgValue = macAlgValue;
	switch( macAlgValue ){
		case MIKEY_PAYLOAD_KEMAC_MAC_HMAC_SHA1_160:
			this->macDataPtr = new byte_t[ 20 ];
			memcpy( this->macDataPtr, macDataPtr, 20 );
			break;
		case MIKEY_PAYLOAD_KEMAC_MAC_NULL:
			this->macDataPtr = NULL;
		default:
			throw new MikeyExceptionMessageContent(
					"Unknown MAC algorithm" );
	}
}

MikeyPayloadKEMAC::MikeyPayloadKEMAC(byte_t *start, int lengthLimit):MikeyPayload(start){
	if(  lengthLimit < 5 ){
		throw new MikeyExceptionMessageLengthException(
                        "Given data is too short to form a KEMAC Payload" );
                return;
        }
	this->payloadTypeValue = MIKEYPAYLOAD_KEMAC_PAYLOAD_TYPE;
	setNextPayloadType(start[0]);
	encrAlgValue = start[1];
	encrDataLengthValue = ( (int)start[2] << 8) |
		             (int)start[3];
	if( lengthLimit < 5 + encrDataLengthValue ){	
		throw new MikeyExceptionMessageLengthException(
                        "Given data is too short to form a KEMAC Payload" );
                return;
        }
	macAlgValue = (int)start[4 + encrDataLengthValue];
	switch( macAlgValue ){
		case MIKEY_PAYLOAD_KEMAC_MAC_HMAC_SHA1_160:
			if( lengthLimit < 25 + encrDataLengthValue ){	
				throw new MikeyExceptionMessageLengthException(
                        	"Given data is too short to form" 
				"a KEMAC Payload" );
              			return;
        		}
			this->macDataPtr = new byte_t[ 20 ];
			memcpy( this->macDataPtr, &
				start[5 + encrDataLengthValue],
				20 );
			endPtr = startPtr + 25 + encrDataLengthValue;
			break;
		case MIKEY_PAYLOAD_KEMAC_MAC_NULL:
			this->macDataPtr = NULL;
			endPtr = startPtr + 5 + encrDataLengthValue;
			break;
		default:
			throw new MikeyExceptionMessageContent(
				"Unknown MAC algorithm in KEYMAC Payload" );
			return;
	}
	
	encrDataPtr = new byte_t[ encrDataLengthValue ];
	memcpy( encrDataPtr, &start[4], encrDataLengthValue );

	assert( endPtr - startPtr == length() );

}

MikeyPayloadKEMAC::~MikeyPayloadKEMAC(){
	if( encrDataPtr != NULL )
		delete [] encrDataPtr;
	if( macDataPtr != NULL )
		delete [] macDataPtr;
}

int MikeyPayloadKEMAC::length(){
	return  5 + encrDataLengthValue + 
		( (macAlgValue == MIKEY_PAYLOAD_KEMAC_MAC_HMAC_SHA1_160)?20:0 );
}

void MikeyPayloadKEMAC::writeData(byte_t *start, int expectedLength){
	assert( expectedLength == length() );
	start[0] = (byte_t)nextPayloadType();
	start[1] = (byte_t)( encrAlgValue & 0xFF );
	start[2] = (byte_t)( (encrDataLengthValue >> 8) & 0xFF );
	start[3] = (byte_t)( (encrDataLengthValue) & 0xFF );
	memcpy( &start[4], encrDataPtr, encrDataLengthValue );
	start[4 + encrDataLengthValue] = (byte_t)( macAlgValue&0xFF );
	if( macAlgValue == MIKEY_PAYLOAD_KEMAC_MAC_HMAC_SHA1_160 )
		memcpy( &start[5 + encrDataLengthValue], macDataPtr, 20 );
}

int MikeyPayloadKEMAC::encrAlg(){
	return encrAlgValue;
}

int MikeyPayloadKEMAC::encrDataLength(){
	return encrDataLengthValue;
}

byte_t * MikeyPayloadKEMAC::encrData(){
	return encrDataPtr;
}

int MikeyPayloadKEMAC::macAlg(){
	return macAlgValue;
}

byte_t * MikeyPayloadKEMAC::macData(){
	return macDataPtr;
}

list<MikeyPayloadKeyData *> MikeyPayloadKEMAC::keyData( 
		byte_t * encrKey, int encrKeyLength,
		byte_t * iv){
	
	byte_t decrData[ encrDataLengthValue ];
	AES * aes;
	int limit = encrDataLengthValue;
	list<MikeyPayloadKeyData *> output;
	MikeyPayloadKeyData * payload;

	switch( encrAlgValue ){
		case MIKEY_PAYLOAD_KEMAC_ENCR_AES_CM_128:
			aes = new AES( encrKey, encrKeyLength );
			aes->ctr_encrypt( encrDataPtr, encrDataLengthValue, 
					decrData, iv );
			delete aes;
			break;
		case MIKEY_PAYLOAD_KEMAC_ENCR_NULL:
			memcpy( decrData, encrDataPtr, encrDataLengthValue );
			break;
		case MIKEY_PAYLOAD_KEMAC_ENCR_AES_KW_128:
			//TODO
		default:
			throw new MikeyException( 
					"Unknown encryption algorithm" );
			break;
	}

	do{
	output.push_back( 
		payload = new MikeyPayloadKeyData( decrData, limit ) );
	limit -= payload->length();
	} while( payload->nextPayloadType() != MIKEYPAYLOAD_LAST_PAYLOAD );

	assert( limit == 0 );
	return output;
}

string MikeyPayloadKEMAC::debugDump(){
	return "MikeyPayloadKEMAC: encrAlgValue=<" + itoa( encrAlgValue ) +
		"> encrDataLengthValue=<" + itoa( encrDataLengthValue ) +
		"> encrDataPtr=<" + 
		print_hex( encrDataPtr, encrDataLengthValue ) +
		"> macAlgValue=<" + itoa( macAlgValue ) +
		"> macDataPtr=<" + print_hex( macDataPtr, 
		((macAlgValue==MIKEY_PAYLOAD_KEMAC_MAC_HMAC_SHA1_160)?20:0)
		) + ">";
}

void MikeyPayloadKEMAC::setMac( byte_t * data ){
	if( macDataPtr != NULL )
		delete [] macDataPtr;

	switch( macAlgValue ){
		case MIKEY_PAYLOAD_KEMAC_MAC_HMAC_SHA1_160:
			macDataPtr = new byte_t[ 20 ];
			memcpy( macDataPtr, data, 20 );
			break;
		case MIKEY_PAYLOAD_KEMAC_MAC_NULL:
			macDataPtr = NULL;
			break;
		default:
			throw new MikeyException( 
				"Unknown MAC algorithm" );
	}
}
